#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct _Summary Summary;
typedef struct _SummaryBlock SummaryBlock;
typedef struct _SummaryItem SummaryItem;

struct _SummaryItem {
	char *uid;
	int seq;
	int flags;
	size_t size;

	union {
		struct {
			off_t from_offset;
			off_t subject_offset;
			off_t to_offset;
			off_t cc_offset;
		} offsets;
		struct {
			char *from;
			char *subject;
			char *to;
			char *cc;
		} pointers;
	} mmappable;

	int mem;
	SummaryBlock *block;
	SummaryItem *next;
};

struct _SummaryBlock {
	int ref_count;
	SummaryItem *items;
	char *data;
	FILE *data_file;
	int mmap_size;
	int block_id;
	int written;
};

struct _Summary {
	GPtrArray *blocks;
};

static char*
summary_block_index_filename (SummaryBlock *b)
{
	return g_strdup_printf ("index_%d.idx", b->block_id);
}

static char*
summary_block_flags_filename (SummaryBlock *b)
{
	return g_strdup_printf ("flags_%d.idx", b->block_id);
}

static char*
summary_block_data_filename (SummaryBlock *b)
{
	return g_strdup_printf ("data_%d.mmap", b->block_id);
}

static SummaryBlock*
create_new_summaryblock (int block_id)
{
	SummaryBlock *b = g_slice_new0 (SummaryBlock);
	b->block_id = block_id;
	b->written = FALSE;
	return b;
}

void
summary_add_item (Summary *s, SummaryItem *i)
{
	int block_id = 0; //i->seq % 1000;
	SummaryBlock *b;
	SummaryItem *cur, *prev;

	if (s->blocks->len < block_id+1) {
		int t;
		for (t = s->blocks->len; t < block_id+1; t++) {
			SummaryBlock *mnew = create_new_summaryblock (t);
			g_ptr_array_add (s->blocks, mnew);
		}
	}

	b = (SummaryBlock *) s->blocks->pdata[block_id];
	cur = b->items;
	if (cur) {
		prev = b->items;
		while (cur) {
			prev = cur;
			cur = cur->next;
		}
		prev->next = i;
	} else {
		b->items = i;
	}

	b->ref_count++;
	b->written = FALSE;
	i->next = NULL;
}

static SummaryBlock *
read_summary_block (int block_id)
{
	FILE *idx, *flags;
	SummaryBlock *b = g_slice_new0 (SummaryBlock);
	struct stat buf;
	SummaryItem *cur = NULL;
	char *index_filename;
	char *data_filename;
	char *flags_filename;

	b->block_id = block_id;
	b->written = TRUE;

	index_filename = summary_block_index_filename (b);
	data_filename = summary_block_data_filename (b);
	flags_filename = summary_block_flags_filename (b);

	idx = fopen (index_filename, "r");
	b->data_file = fopen (data_filename, "r");
	flags = fopen (flags_filename, "r");

	g_free (index_filename);
	g_free (data_filename);
	g_free (flags_filename);

	fstat(fileno (b->data_file), &buf);
	b->mmap_size = buf.st_size;
	b->data = mmap (NULL, b->mmap_size, PROT_READ, MAP_PRIVATE, fileno (b->data_file), 0);

	while (!feof (idx)) {
		int len;
		if (fscanf (idx, "%d", &len) != EOF) 
		{
			SummaryItem *ncur = g_slice_new0 (SummaryItem);
			ncur->block = b;
			ncur->block->ref_count++;
			ncur->next = NULL;

			if (!b->items)
				b->items = ncur;
			else
				cur->next = ncur;

			cur = ncur;

			/* Slab alloc here (most will have equal size, frequent kind) */
			cur->mem = 0;
			cur->uid = (char *) malloc(len);
			memset (cur->uid, 0, len);
			fgetc (idx);
			fread (cur->uid, 1, len, idx);
			cur->uid[len] = '\0';
			fscanf (idx, "%d%d%d%d%d%d",
				&cur->seq, &cur->size,
				(int *) &cur->mmappable.offsets.subject_offset, 
				(int *) &cur->mmappable.offsets.from_offset,
				(int *) &cur->mmappable.offsets.to_offset,
				(int *) &cur->mmappable.offsets.cc_offset);

		}
	}	
	fclose (idx);

	/* Perhaps scan this in a table and loop the table in above loop  in stead */
	while (!feof (flags)) {
		SummaryItem *c = b->items;
		int seq, flag;
		if (fscanf (flags, "%d%d", &seq, &flag) != EOF) {
			while (c) {
				if (c->seq == seq) {
					c->flags = flag;
				}
				c = c->next;
			}
		}
	}
	fclose(flags);

	return b;
}

const char*
summary_item_get_uid (SummaryItem *i)
{
	return i->uid;
}

int
summary_item_get_seq (SummaryItem *i)
{
	return (const int) i->seq;
}

int
summary_item_get_flags (SummaryItem *i)
{
	return (const int) i->flags;
}

void
summary_item_set_flags (SummaryItem *i, int new_flags)
{
	i->flags = new_flags;
	/* todo: rewrite flags file */
	return;
}


const char*
summary_item_get_subject (SummaryItem *i)
{
	if (i->mem)
		return i->mmappable.pointers.subject;

	return i->block->data + i->mmappable.offsets.subject_offset;
}

const char*
summary_item_get_from (SummaryItem *i)
{
	if (i->mem)
		return i->mmappable.pointers.from;

	return i->block->data + i->mmappable.offsets.from_offset;
}


const char*
summary_item_get_to (SummaryItem *i)
{
	if (i->mem)
		return i->mmappable.pointers.to;

	return i->block->data + i->mmappable.offsets.to_offset;
}

const char*
summary_item_get_cc (SummaryItem *i)
{
	if (i->mem)
		return i->mmappable.pointers.cc;
 
	return i->block->data + i->mmappable.offsets.cc_offset;
}


static void
summary_item_free (SummaryItem *i)
{
	if (i->block)
		i->block->ref_count--;
	g_free (i->uid);
	if (i->mem) {
		g_free (i->mmappable.pointers.from);
		g_free (i->mmappable.pointers.subject);
		g_free (i->mmappable.pointers.to);
		g_free (i->mmappable.pointers.cc);
		i->mem = 0;
	}
	g_slice_free (SummaryItem, i);
}


typedef struct {
	char *str;
	int cnt, cpy;
	GList *affected;
} StoreString;

static gint 
find_storestring (gconstpointer a, gconstpointer b)
{
	StoreString *sa = (StoreString *) a;
	char *cb = (char *) b;
	if (!strcmp (sa->str, cb))
		return 0;
	return 1;
}

static gint 
cmp_storestring (gconstpointer a, gconstpointer b)
{
	StoreString *sa = (StoreString *) a;
	StoreString *sb = (StoreString *) b;
	return sb->cnt - sa->cnt;
}


static void
summary_block_dump (SummaryBlock *b)
{
	FILE *data, *idx, *flags;
	GList *strings = NULL;
	GList *listnode = NULL;
	SummaryItem *item = b->items;

	char *index_filename;
	char *data_filename;
	char *flags_filename;
	char *index_filename_n;
	char *data_filename_n;
	char *flags_filename_n;

	if (b->written)
		return;

	index_filename = summary_block_index_filename (b);
	data_filename = summary_block_data_filename (b);
	flags_filename = summary_block_flags_filename (b);

	index_filename_n = g_strdup_printf ("%s~", index_filename);
	data_filename_n = g_strdup_printf ("%s~", data_filename);
	flags_filename_n = g_strdup_printf ("%s~", flags_filename);

	idx = fopen (index_filename_n, "w");
	data = fopen (data_filename_n, "w");
	flags = fopen (flags_filename_n, "w");


	while (item) {
		StoreString *sb;
		char *str;
		gboolean cpy = FALSE;

		/* subject */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.subject_offset;
		else {
			str = g_strdup (item->mmappable.pointers.subject);
			cpy = TRUE;
		}

		listnode = g_list_find_custom (strings, str, find_storestring);
		if (listnode) {
			sb = (StoreString *) listnode->data;
			sb->cnt++;
			if (cpy)
				g_free (str);
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = str;
			sb->cpy = cpy;
			strings = g_list_prepend (strings, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.subject_offset);
		cpy = FALSE;

		/* from */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.from_offset;
		else {
			str = g_strdup (item->mmappable.pointers.from);
			cpy = TRUE;
		}

		listnode = g_list_find_custom (strings, str, find_storestring);
		if (listnode) {
			sb = (StoreString *) listnode->data;
			sb->cnt++;
			if (cpy)
				g_free (str);
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = str;
			sb->cpy = cpy;
			strings = g_list_prepend (strings, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.from_offset);
		cpy = FALSE;


		/* to */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.to_offset;
		else {
			str = g_strdup (item->mmappable.pointers.to);
			cpy = TRUE;
		}

		listnode = g_list_find_custom (strings, str, find_storestring);
		if (listnode) {
			sb = (StoreString *) listnode->data;
			sb->cnt++;
			if (cpy)
				g_free (str);
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = str;
			sb->cpy = cpy;
			strings = g_list_prepend (strings, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.to_offset);
		cpy = FALSE;


		/* cc */
		if (!item->mem)
			str = item->block->data + item->mmappable.offsets.cc_offset;
		else {
			str = g_strdup (item->mmappable.pointers.cc);
			cpy = TRUE;
		}

		listnode = g_list_find_custom (strings, str, find_storestring);
		if (listnode) {
			sb = (StoreString *) listnode->data;
			sb->cnt++;
			if (cpy)
				g_free (str);
			sb->cpy = FALSE;
		} else {
			sb = g_slice_new0 (StoreString);
			sb->cnt = 1;
			sb->str = str;
			sb->cpy = cpy;
			strings = g_list_prepend (strings, sb);
		}
		sb->affected = g_list_prepend (sb->affected, &item->mmappable.offsets.cc_offset);
		cpy = FALSE;

		/* After this we will be a memory one (this needs a lock) */

		if (item->mem) {
			g_free (item->mmappable.pointers.from);
			g_free (item->mmappable.pointers.subject);
			g_free (item->mmappable.pointers.to);
			g_free (item->mmappable.pointers.cc);
			item->mem = 0;
		}

		item = item->next;
	}

	strings = g_list_sort (strings, cmp_storestring);

	
	while (strings) {
		StoreString *ss = (StoreString *) strings->data;
		int location = ftell (data);

		printf ("%s %d times\n", ss->str, ss->cnt);

		fputs (ss->str, data);
		if (ss->cpy) {
			g_free (ss->str);
			ss->cpy = FALSE;
		}
		fputc ('\0', data);

		while (ss->affected) {
			int *data = ss->affected->data;
			*data = location;
			ss->affected = g_list_next (ss->affected);
		}
		g_list_free (ss->affected);
		
		g_slice_free (StoreString, ss);
		strings = g_list_next (strings);
	}

	g_list_free (strings);

	fclose (data);

	item = b->items;
	while (item) {

		fprintf (flags, "%d %d\n", item->seq, item->flags); 
		fprintf (idx, "%d %s %d %d %d %d %d %d\n",
				strlen (item->uid), item->uid,
				item->seq, item->size,
				(int) item->mmappable.offsets.subject_offset, 
				(int) item->mmappable.offsets.from_offset,
				(int) item->mmappable.offsets.to_offset,
				(int) item->mmappable.offsets.cc_offset);

		item = item->next;
	}

	fclose (idx);
	fclose (flags);

	rename (index_filename_n, index_filename);
	rename (data_filename_n, data_filename);
	rename (flags_filename_n, flags_filename);

	g_free (index_filename);
	g_free (data_filename);
	g_free (flags_filename);
	g_free (index_filename_n);
	g_free (data_filename_n);
	g_free (flags_filename_n);

}

static void
summary_dump (Summary *summary)
{
	int i;
	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		summary_block_dump (b);
	}
}

static void
summary_block_free (SummaryBlock *b)
{
	SummaryItem *i = b->items;
	while (i) {
		summary_item_free (i);
		i = i->next;
	}
	if (b->ref_count <= 0) {
		if (b->data)
			munmap (b->data, b->mmap_size);
		if (b->data_file)
			fclose (b->data_file);
		g_slice_free (SummaryBlock, b);
	}
}

static void
summary_free (Summary *summary)
{
	int i;
	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		summary_block_free (b);
	}
	g_ptr_array_free (summary->blocks, TRUE);
	g_slice_free (Summary, summary);
}

static Summary*
read_summary (void)
{
	Summary *summary = g_slice_new0 (Summary);
	int i, items;
	FILE *f = fopen ("summary.idx", "r");

	fscanf (f, "%d", &items);

	summary->blocks = g_ptr_array_sized_new (items);

	for (i=0; i<items; i++) {
		SummaryBlock *b = read_summary_block (i);
		g_ptr_array_add (summary->blocks, b);		
	}

	return summary;
}

SummaryItem*
create_summary_item (const char *uid, int seq, int flags, size_t size, const char *from, const char *subject, const char *to, const char *cc)
{
	SummaryItem *item = g_slice_new0 (SummaryItem);

	item->size = size;
	item->seq = seq;
	item->flags = flags;
	item->uid = g_strdup (uid);
	item->mmappable.pointers.from = g_strdup (from);
	item->mmappable.pointers.to = g_strdup (to);
	item->mmappable.pointers.cc = g_strdup (cc);
	item->mmappable.pointers.subject = g_strdup (subject);
	item->next = NULL;
	item->block = NULL;
	item->mem = 1; /* memory item */

	return item;
}

int main()
{
	Summary *summary = read_summary ();
	int i;

	for (i = 0; i < summary->blocks->len; i++) {
		SummaryBlock *b = (SummaryBlock *) summary->blocks->pdata[i];
		SummaryItem *mc = b->items;
		
		while (mc) {
			printf ("UID %s is at sequence %d with flags %d\n", 
				summary_item_get_uid (mc),
				summary_item_get_seq (mc),
				summary_item_get_flags (mc));

			printf ("\tFrom=%s\n\tSubj=%s\n\tCc=%s\n\tTo=%s\n",
				summary_item_get_from (mc),
				summary_item_get_subject (mc),
				summary_item_get_cc (mc),
				summary_item_get_to (mc));
	
			mc = mc->next;
		}
	}

	for (i = 0; i < 5; i++) {	
		const char *uid = g_strdup_printf ("uid%d", i);
		int seq = 10+i;
		int flags = i+1;
		size_t size = i+2;
		char *from = g_strdup_printf ("from%d", i);
		const char *subject = g_strdup_printf ("subj%d", i);
		const char *to = g_strdup_printf ("to%d", i);
		const char *cc = g_strdup_printf ("cc%d", i);

		SummaryItem *item = create_summary_item (uid, seq, flags, size, from, subject, to, cc);
	
		summary_add_item (summary, item);
	}

	summary_dump (summary);
	summary_free (summary);

	return 0;
}

